# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os

# project
from ...defaults import Defaults
from ...system.prepare import SystemPrepare
from ...system.profile import Profile
from ...utils.compress import Compress
from ...system.setup import SystemSetup
from ...logger import log
from ...command import Command
from ...system.kernel import Kernel
from .base import BootImageBase


class BootImageDracut(BootImageBase):
    """
    Implements preparation and creation of kiwi boot(initrd) images
    The kiwi initrd is a customized first boot initrd which allows
    to control the first boot an appliance. The kiwi initrd replaces
    itself after first boot by the result of dracut.
    """
    def prepare(self):
        """
        Prepare new root system suitable to run dracut as chrooted
        operation to create the initrd
        """
        self.load_boot_xml_description()
        boot_image_name = self.boot_xml_state.xml_data.get_name()

        self.import_system_description_elements()

        log.info('Preparing boot image')
        system = SystemPrepare(
            xml_state=self.boot_xml_state,
            root_dir=self.boot_root_directory,
            allow_existing=True
        )
        manager = system.setup_repositories()
        system.install_bootstrap(
            manager
        )
        system.install_system(
            manager
        )

        profile = Profile(self.boot_xml_state)
        profile.add('kiwi_initrdname', boot_image_name)

        defaults = Defaults()
        defaults.to_profile(profile)

        self.setup = SystemSetup(
            self.boot_xml_state, self.boot_root_directory
        )
        self.setup.import_shell_environment(profile)
        self.setup.import_description()
        self.setup.call_image_script()

    def create_initrd(self, mbrid=None):
        """
        Call dracut to create the initrd and XZ compress the result

        :param object mbrid: unused
        """
        if self.is_prepared():
            log.info('Creating generic dracut initrd archive')
            kernel_info = Kernel(self.boot_root_directory)
            kernel_version = kernel_info.get_kernel().version
            Command.run(
                [
                    'chroot', self.boot_root_directory,
                    'dracut', '--force',
                    '--no-hostonly',
                    '--no-hostonly-cmdline',
                    '--no-compress',
                    os.path.basename(self.initrd_file_name),
                    kernel_version
                ]
            )
            Command.run(
                [
                    'mv',
                    self.boot_root_directory + '/' + os.path.basename(
                        self.initrd_file_name
                    ),
                    self.initrd_file_name
                ]
            )
            log.info(
                '--> xz compressing archive'
            )
            compress = Compress(self.initrd_file_name)
            compress.xz()
            self.initrd_filename = compress.compressed_filename



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
# project
from .qcow2 import DiskFormatQcow2
from .vhd import DiskFormatVhd
from .vhdfixed import DiskFormatVhdFixed
from .vmdk import DiskFormatVmdk
from .gce import DiskFormatGce
from .vdi import DiskFormatVdi

from ...exceptions import (
    KiwiDiskFormatSetupError
)


class DiskFormat(object):
    """
    DiskFormat factory

    Attributes

    * :attr:`name`
        Format name

    * :attr:`xml_state`
        Instance of XMLState

    * :attr:`root_dir`
        root directory path name

    * :attr:`target_dir`
        target directory path name
    """
    def __new__(self, name, xml_state, root_dir, target_dir):
        if name == 'qcow2':
            return DiskFormatQcow2(
                xml_state, root_dir, target_dir
            )
        elif name == 'vdi':
            return DiskFormatVdi(
                xml_state, root_dir, target_dir
            )
        elif name == 'vhd':
            return DiskFormatVhd(
                xml_state, root_dir, target_dir
            )
        elif name == 'vhdfixed':
            custom_args = None
            disk_tag = xml_state.build_type.get_vhdfixedtag()
            if disk_tag:
                custom_args = {
                    '--tag': disk_tag
                }
            return DiskFormatVhdFixed(
                xml_state, root_dir, target_dir, custom_args
            )
        elif name == 'gce':
            custom_args = None
            gce_license_tag = xml_state.build_type.get_gcelicense()
            if gce_license_tag:
                custom_args = {
                    '--tag': gce_license_tag
                }
            return DiskFormatGce(
                xml_state, root_dir, target_dir, custom_args
            )
        elif name == 'vmdk':
            custom_args = None
            vmdisk_section = xml_state.get_build_type_vmdisk_section()
            if vmdisk_section:
                custom_args = {}
                disk_mode = vmdisk_section.get_diskmode()
                disk_controller = vmdisk_section.get_controller()
                if disk_mode:
                    custom_args['subformat=%s' % disk_mode] = None
                if disk_controller:
                    custom_args['adapter_type=%s' % disk_controller] = None
            return DiskFormatVmdk(
                xml_state, root_dir, target_dir, custom_args
            )
        else:
            raise KiwiDiskFormatSetupError(
                'Support for %s disk format not implemented' % name
            )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import stat
from pwd import getpwnam
from tempfile import mkdtemp
from shutil import rmtree
import os

# project
from ..utils.sync import DataSync
from ..command import Command
from ..path import Path

from ..exceptions import (
    KiwiRootDirExists,
    KiwiRootInitCreationError
)


class RootInit(object):
    """
    Implements creation of new root directory for a linux system.
    Host system independent static default files and device nodes
    are created to initialize a new base system

    Attributes

    * :attr:`root_dir`
        root directory path name
    """
    def __init__(self, root_dir, allow_existing=False):
        if not allow_existing and os.path.exists(root_dir):
            raise KiwiRootDirExists(
                'Root directory %s already exists' % root_dir
            )
        self.root_dir = root_dir

    def delete(self):
        """
        Force delete root directory and its contents
        """
        Path.wipe(self.root_dir)

    def create(self):
        """
        Create new system root directory

        The method creates a temporary directory and initializes it
        for the purpose of building a system image from it. This
        includes the following setup:

        * create static core device nodes
        * create core system paths

        On success the contents of the temporary location are
        synced to the specified root_dir and the temporary location
        will be deleted. That way we never work on an incomplete
        initial setup
        """
        root = mkdtemp(prefix='kiwi_root.')
        try:
            self._create_base_directories(root)
            self._create_device_nodes(root)
            self._create_base_links(root)
            self._setup_config_templates(root)
            data = DataSync(root + '/', self.root_dir)
            data.sync_data(
                options=['-a', '--ignore-existing']
            )
            rmtree(root, ignore_errors=True)
        except Exception as e:
            rmtree(root, ignore_errors=True)
            raise KiwiRootInitCreationError(
                '%s: %s' % (type(e).__name__, format(e))
            )

    def _setup_config_templates(self, root):
        group_template = '/var/adm/fillup-templates/group.aaa_base'
        passwd_template = '/var/adm/fillup-templates/passwd.aaa_base'
        proxy_template = '/var/adm/fillup-templates/sysconfig.proxy'
        if os.path.exists(group_template):
            Command.run(['cp', group_template, root + '/etc/group'])
        if os.path.exists(passwd_template):
            Command.run(['cp', passwd_template, root + '/etc/passwd'])
        if os.path.exists(proxy_template):
            Command.run(['cp', proxy_template, root + '/etc/sysconfig/proxy'])

    def _create_device_nodes(self, root):
        """
        Create minimum set of device nodes for a new root system
        """
        mknod_data = [
            (0o666, 'dev/null', stat.S_IFCHR, (1, 3)),
            (0o666, 'dev/zero', stat.S_IFCHR, (1, 5)),
            (0o622, 'dev/full', stat.S_IFCHR, (1, 7)),
            (0o666, 'dev/random', stat.S_IFCHR, (1, 8)),
            (0o644, 'dev/urandom', stat.S_IFCHR, (1, 9)),
            (0o666, 'dev/tty', stat.S_IFCHR, (5, 0)),
            (0o666, 'dev/ptmx', stat.S_IFCHR, (5, 2)),
            (0o640, 'dev/loop0', stat.S_IFBLK, (7, 0)),
            (0o640, 'dev/loop1', stat.S_IFBLK, (7, 1)),
            (0o640, 'dev/loop2', stat.S_IFBLK, (7, 2)),
            (0o666, 'dev/loop3', stat.S_IFBLK, (7, 3)),
            (0o666, 'dev/loop4', stat.S_IFBLK, (7, 4))
        ]
        for d_mode, d_path, d_type, d_dev in mknod_data:
            root_path = os.sep.join([root, d_path])
            os.mknod(root_path, d_mode | d_type, os.makedev(*d_dev))

    def _create_base_directories(self, root):
        """
        Create minimum collection of directories in the new
        root system required for kiwi to operate
        """
        base_system_paths = (
            'dev/pts',
            'proc',
            'etc/sysconfig',
            'var/cache',
            'run',
            'sys'
        )
        root_uid = getpwnam('root').pw_uid
        root_gid = getpwnam('root').pw_gid
        for path in base_system_paths:
            root_path = os.sep.join([root, path])
            os.makedirs(root_path)
            os.chown(root_path, root_uid, root_gid)

    def _create_base_links(self, root):
        """
        Create minimum collection of file handle and runtime
        links which needs to be present prior to any package
        installation
        """
        base_system_links = (
            ('/proc/self/fd', '%s/dev/fd'),
            ('fd/2', '%s/dev/stderr'),
            ('fd/0', '%s/dev/stdin'),
            ('fd/1', '%s/dev/stdout'),
            ('/run', '%s/var/run')
        )
        for src, target in base_system_links:
            os.symlink(src, target % root, )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from tempfile import mkdtemp

from ...defaults import Defaults
from ...utils.sync import DataSync
from ...system.prepare import SystemPrepare
from ...system.profile import Profile
from ...system.setup import SystemSetup
from ...logger import log
from ...archive.cpio import ArchiveCpio
from ...utils.compress import Compress
from ...path import Path
from .base import BootImageBase


class BootImageKiwi(BootImageBase):
    """
    Implements preparation and creation of kiwi boot(initrd) images
    The kiwi initrd is a customized first boot initrd which allows
    to control the first boot an appliance. The kiwi initrd replaces
    itself after first boot by the result of dracut.
    """
    def prepare(self):
        """
        Prepare new root system suitable to create a kiwi initrd from it
        """
        self.load_boot_xml_description()
        boot_image_name = self.boot_xml_state.xml_data.get_name()

        self.import_system_description_elements()

        log.info('Preparing boot image')
        system = SystemPrepare(
            xml_state=self.boot_xml_state,
            root_dir=self.boot_root_directory,
            allow_existing=True
        )
        manager = system.setup_repositories()
        system.install_bootstrap(
            manager
        )
        system.install_system(
            manager
        )

        profile = Profile(self.boot_xml_state)
        profile.add('kiwi_initrdname', boot_image_name)

        defaults = Defaults()
        defaults.to_profile(profile)

        self.setup = SystemSetup(
            self.boot_xml_state, self.boot_root_directory
        )
        self.setup.import_shell_environment(profile)
        self.setup.import_description()
        self.setup.import_overlay_files(
            follow_links=True
        )
        self.setup.call_config_script()

        system.pinch_system(
            manager=manager, force=True
        )

        self.setup.call_image_script()
        self.setup.create_init_link_from_linuxrc()

    def create_initrd(self, mbrid=None):
        """
        Create initrd from prepared boot system tree and compress the result

        :param object mbrid: instance of ImageIdentifier
        """
        if self.is_prepared():
            log.info('Creating initrd cpio archive')
            # we can't simply exclude boot when building the archive
            # because the file boot/mbrid must be preserved. Because of
            # that we create a copy of the boot directory and remove
            # everything in boot/ except for boot/mbrid. The original
            # boot directory should not be changed because we rely
            # on other data in boot/ e.g the kernel to be available
            # for the entire image building process
            temp_boot_root_directory = mkdtemp(
                prefix='kiwi_boot_root_copy.'
            )
            self.temp_directories.append(
                temp_boot_root_directory
            )
            data = DataSync(
                self.boot_root_directory + '/',
                temp_boot_root_directory
            )
            data.sync_data(
                options=['-z', '-a']
            )
            boot_directory = temp_boot_root_directory + '/boot'
            Path.wipe(boot_directory)
            if mbrid:
                log.info(
                    '--> Importing mbrid: %s', mbrid.get_id()
                )
                Path.create(boot_directory)
                image_identifier = boot_directory + '/mbrid'
                mbrid.write(image_identifier)

            cpio = ArchiveCpio(self.initrd_file_name)
            # the following is a list of directories which were needed
            # during the process of creating an image but not when the
            # image is actually booting with this initrd
            exclude_from_archive = [
                '/var/cache', '/image', '/usr/lib/grub*'
            ]
            # the following is a list of directories to exclude which
            # are not needed inside of the initrd
            exclude_from_archive += [
                '/usr/share/doc', '/usr/share/man', '/home', '/media', '/srv'
            ]
            cpio.create(
                source_dir=temp_boot_root_directory,
                exclude=exclude_from_archive
            )
            log.info(
                '--> xz compressing archive'
            )
            compress = Compress(self.initrd_file_name)
            compress.xz()
            self.initrd_filename = compress.compressed_filename



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import platform
from collections import OrderedDict

from ...exceptions import (
    KiwiFormatSetupError
)

from ...defaults import Defaults
from ...path import Path
from ...logger import log


class DiskFormatBase(object):
    """
    Base class to create disk formats from a raw disk image

    Attributes

    * :attr:`xml_state`
        Instance of XMLState

    * :attr:`root_dir`
        root directory path name

    * :attr:`arch`
        platform.machine

    * :attr:`target_dir`
        target directory path name

    * :attr:`custom_args`
        list of custom format options

    * :attr:`temp_image_dir`
        temporary manifest directory

    * :attr:`diskname`
        raw disk file path name

    * :attr:`image_format`
        disk format name
    """
    def __init__(self, xml_state, root_dir, target_dir, custom_args=None):
        self.xml_state = xml_state
        self.root_dir = root_dir
        self.arch = platform.machine()
        self.target_dir = target_dir
        self.custom_args = {}
        self.temp_image_dir = None
        self.image_format = None
        self.diskname = self.get_target_name_for_format('raw')

        self.post_init(custom_args)

    def post_init(self, custom_args):
        """
        Post initialization method

        Implementation in specialized disk format class if required

        :param list custom_args: unused
        """
        pass

    def create_image_format(self):
        """
        Create disk format

        Implementation in specialized disk format class required
        """
        raise NotImplementedError

    def get_qemu_option_list(self, custom_args):
        """
        Create list of qemu options from custom_args dict

        :param dict custom_args: arguments

        :return: qemu option list
        :rtype: list
        """
        options = []
        if custom_args:
            ordered_args = OrderedDict(sorted(custom_args.items()))
            for key, value in list(ordered_args.items()):
                options.append('-o')
                options.append(key)
                if value:
                    options.append(value)
        return options

    def get_target_name_for_format(self, format_name):
        """
        Create target file path name for specified format

        :param string format_name: disk format name

        :return: file path name
        :rtype: string
        """
        if format_name != 'raw':
            if format_name not in Defaults.get_disk_format_types():
                raise KiwiFormatSetupError(
                    'unsupported disk format %s' % format_name
                )
        return ''.join(
            [
                self.target_dir, '/',
                self.xml_state.xml_data.get_name(),
                '.' + self.arch,
                '-' + self.xml_state.get_image_version(),
                '.' + format_name
            ]
        )

    def store_to_result(self, result):
        """
        Store result file of the format conversion into the
        provided result instance.

        By default only the converted image file will be stored.
        Subformats which creates additional metadata files needs
        to overwrite this method

        :param object result: Instance of Result
        """
        result.add(
            key='disk_format_image',
            filename=self.get_target_name_for_format(
                self.image_format
            ),
            use_for_bundle=True,
            compress=False,
            shasum=True
        )

    def __del__(self):
        if self.temp_image_dir and os.path.exists(self.temp_image_dir):
            log.info('Cleaning up %s instance', type(self).__name__)
            Path.wipe(self.temp_image_dir)
